"""Database operations."""
from datetime import datetime, timedelta
from typing import Optional, List
import databases
from dataclasses import dataclass


@dataclass(frozen=True)
class AveragedRead:
    """An averaged read result.

    Includes the number of reads that went into the average, as well as the oldest timestamp of those reads.
    """

    avg_pm25: float
    avg_pm10: float
    count: int
    oldest_read_time: datetime


@dataclass(frozen=True)
class ReadLogEntry:
    """A read log entry."""

    event_time: datetime
    pm25: float
    pm10: float


@dataclass(frozen=True)
class EpaAqiLogEntry:
    """An EPA Aqi log entry."""

    event_time: datetime
    epa_aqi: float
    pollutant: str
    read_count: int
    oldest_read_time: datetime


async def get_latest_read(dbconn: databases.Database) -> ReadLogEntry:
    """Get the most recent read from the database."""
    result = await dbconn.fetch_one("SELECT event_time, pm25, pm10 FROM read_log ORDER BY event_time DESC LIMIT 1")
    if result:
        return ReadLogEntry(datetime.fromtimestamp(result[0]), result[1], result[2])
    else:
        return ReadLogEntry(datetime.now(), 0, 0)


async def get_latest_epa_aqi(dbconn: databases.Database) -> EpaAqiLogEntry:
    """Get the most recent EPA AQI from the database."""
    result = await dbconn.fetch_one(
        "SELECT event_time, epa_aqi, pollutant, read_count, oldest_read_time "
        "FROM epa_aqi_log ORDER BY event_time DESC LIMIT 1"
    )
    if result:
        return EpaAqiLogEntry(
            event_time=datetime.fromtimestamp(result[0]),
            epa_aqi=result[1],
            pollutant=result[2],
            read_count=result[3],
            oldest_read_time=datetime.fromtimestamp(result[4]),
        )
    else:
        return EpaAqiLogEntry(
            event_time=datetime.now(), epa_aqi=0, pollutant="NA", read_count=0, oldest_read_time=datetime.now()
        )


async def get_all_reads(dbconn: databases.Database, lookback: Optional[datetime]) -> List[ReadLogEntry]:
    """Retrieve all read stats, for a given time window.

    If no window is specified, all results are returned.
    """
    if lookback:
        data = await dbconn.fetch_all(
            "SELECT event_time, pm10, pm25 FROM read_log WHERE event_time >= :lookback ORDER BY event_time ASC",
            values={"lookback": int(lookback.timestamp())},
        )
    else:
        data = await dbconn.fetch_all("SELECT event_time, pm10, pm25 FROM read_log ORDER BY event_time ASC")

    return [ReadLogEntry(event_time=datetime.fromtimestamp(x[0]), pm10=x[1], pm25=x[2]) for x in data]


async def get_all_epa_aqis(dbconn: databases.Database, lookback: Optional[datetime]) -> List[EpaAqiLogEntry]:
    """Retrieve all read stats, for a given time window.

    If no window is specified, all results are returned.
    """
    if lookback:
        data = await dbconn.fetch_all(
            "SELECT event_time, epa_aqi, pollutant, read_count, oldest_read_time "
            "FROM epa_aqi_log "
            "WHERE event_time >= :lookback ORDER BY event_time ASC",
            values={"lookback": int(lookback.timestamp())},
        )
    else:
        data = await dbconn.fetch_all(
            "SELECT event_time, epa_aqi, pollutant, read_count, oldest_read_time "
            "FROM epa_aqi_log ORDER BY event_time ASC"
        )

    return [
        EpaAqiLogEntry(
            event_time=datetime.fromtimestamp(x[0]),
            epa_aqi=x[1],
            pollutant=x[2],
            read_count=x[3],
            oldest_read_time=datetime.fromtimestamp(x[4]),
        )
        for x in data
    ]


async def get_averaged_reads(dbconn: databases.Database, lookback_to: datetime) -> Optional[AveragedRead]:
    """Get the average read values, looking back to a certain time.

    Note that the lookback will include one additional value outside of the window if it exists.  This allows for us to
    ensure full coverage of the lookback window.
    """
    lookback = int(lookback_to.timestamp())
    result = await dbconn.fetch_one(
        "SELECT "
        "AVG(pm25) as avg_pm25, AVG(pm10) as avg_pm10, COUNT(*) as count, MIN(event_time) as oldest_time "
        "FROM read_log "
        "WHERE (event_time >= :lookback) OR "
        "(event_time = (SELECT MAX(event_time) FROM read_log WHERE event_time <= :lookback)) ORDER BY event_time ASC",
        values={"lookback": lookback},
    )

    if result is None:
        return None
    else:
        return AveragedRead(
            avg_pm25=result[0],
            avg_pm10=result[1],
            count=result[2],
            oldest_read_time=datetime.fromtimestamp(result[3]),
        )


async def clean_old(dbconn: databases.Database, retention_minutes: int) -> None:
    """Remove expired database entries.

    This is used to keep the database from going infinitely, and allows us to define a retention period.
    """
    last_week = datetime.now() - timedelta(minutes=retention_minutes)
    last_week_timestamp = int(last_week.timestamp())
    await dbconn.execute(
        "DELETE FROM read_log WHERE event_time < :last_week_timestamp",
        values={"last_week_timestamp": last_week_timestamp},
    )


async def add_epa_read(
    dbconn: databases.Database,
    event_time: datetime,
    epa_aqi: float,
    pollutant: str,
    read_count: int,
    oldest_read_time: datetime,
):
    """Add an EPA read entry to the database."""
    formatted_time = int(event_time.timestamp())
    formatted_oldest_read_time = int(oldest_read_time.timestamp())
    await dbconn.execute(
        query="INSERT INTO epa_aqi_log VALUES (:formatted_time, :epa_aqi, :pollutant, :read_count, :oldest_read_time)",
        values={
            "formatted_time": formatted_time,
            "epa_aqi": epa_aqi,
            "pollutant": pollutant,
            "read_count": read_count,
            "oldest_read_time": formatted_oldest_read_time,
        },
    )


async def add_read(dbconn: databases.Database, event_time: datetime, pm25: float, pm10: float):
    """Add a raw read entry to the database."""
    formatted_time = int(event_time.timestamp())
    await dbconn.execute(
        query="INSERT INTO read_log VALUES (:formatted_time, :pm25, :pm10)",
        values={
            "formatted_time": formatted_time,
            "pm25": pm25,
            "pm10": pm10,
        },
    )


async def create_tables(dbconn: databases.Database):
    """Create database tables, if they don't already exist."""
    await dbconn.execute("""CREATE TABLE IF NOT EXISTS read_log (event_time integer, pm25 real, pm10 real)""")
    await dbconn.execute("""CREATE INDEX IF NOT EXISTS read_eventtime ON read_log (event_time)""")
    await dbconn.execute(
        """CREATE TABLE IF NOT EXISTS epa_aqi_log 
        (event_time integer, epa_aqi real, pollutant text, read_count integer, oldest_read_time integer)"""
    )
    await dbconn.execute("""CREATE INDEX IF NOT EXISTS eqpaqi_eventtime ON epa_aqi_log (event_time)""")



"""All exception classes for the SDS011."""


class Sds011Exception(Exception):
    """Base exception for SDS011 device."""

    pass


class ChecksumFailedException(Sds011Exception):
    """Thrown if the checksum value in a response is incorrect."""

    def __init__(self, expected: int, actual: int):
        """Create exception."""
        super().__init__()
        self.expected = expected
        self.actual = actual


class IncorrectCommandException(Sds011Exception):
    """Thrown if the command ID in a response is incorrect."""

    def __init__(self, expected: int, actual: int):
        """Create exception."""
        super().__init__(f"Expected command {expected}, found {actual}")
        self.expected = expected
        self.actual = actual


class IncorrectCommandCodeException(Sds011Exception):
    """Thrown if the command code in a response is incorrect."""

    def __init__(self, expected: int, actual: int):
        """Create exception."""
        super().__init__(f"Expected code {expected}, found {actual}")
        self.expected = expected
        self.actual = actual


class IncorrectWrapperException(Sds011Exception):
    """Thrown if the wrapper of a response (either HEAD or TAIL) is incorrect."""

    pass


class IncompleteReadException(Sds011Exception):
    """Thrown if the device didn't return complete data when asking for a response."""

    pass


class InvalidDeviceIdException(Sds011Exception):
    """Thrown if the trying to set the device ID on an invalid device."""

    pass



"""Config module."""
import os.path
from dataclasses import dataclass

DEFAULT_DB_PATH = "~/.aqimon/db.sqlite"


@dataclass(frozen=True)
class Config:
    """Config data for the application."""

    database_path: str
    poll_frequency_sec: int
    retention_minutes: int

    reader_type: str

    # Nova PM properties
    usb_path: str
    sleep_sec_between_reads: int
    sample_count_per_read: int
    warm_up_sec: int
    command_wait_time: int

    # EPA Properties
    epa_lookback_minutes: int

    # Server properties
    server_port: int
    server_host: str


DEFAULT_CONFIG = Config(
    database_path=os.path.expanduser(DEFAULT_DB_PATH),
    poll_frequency_sec=60 * 15,  # Every 15 minutes
    retention_minutes=60 * 24 * 7,  # 1 week
    epa_lookback_minutes=60 * 8,  # 8 hours
    reader_type="NOVAPM",
    usb_path="/dev/ttyUSB0",
    sleep_sec_between_reads=5,
    warm_up_sec=15,
    command_wait_time=1,
    sample_count_per_read=5,
    server_port=8000,
    server_host="0.0.0.0",
)


def get_config_from_env() -> Config:
    """Get the config from environment variables."""
    return Config(
        database_path=os.path.expanduser(os.environ.get("AQIMON_DB_PATH", DEFAULT_CONFIG.database_path)),
        poll_frequency_sec=int(os.environ.get("AQIMON_POLL_FREQUENCY_SEC", DEFAULT_CONFIG.poll_frequency_sec)),
        retention_minutes=int(os.environ.get("AQIMON_RETENTION_MINUTES", DEFAULT_CONFIG.retention_minutes)),
        reader_type=os.environ.get("AQIMON_READER_TYPE", DEFAULT_CONFIG.reader_type),
        usb_path=os.environ.get("AQIMON_USB_PATH", DEFAULT_CONFIG.usb_path),
        sleep_sec_between_reads=int(
            os.environ.get("AQIMON_SLEEP_SEC_BETWEEN_READS", DEFAULT_CONFIG.sleep_sec_between_reads)
        ),
        warm_up_sec=int(os.environ.get("AQIMON_WARM_UP_SEC", DEFAULT_CONFIG.warm_up_sec)),
        command_wait_time=int(os.environ.get("AQIMON_COMMAND_WAIT_TIME", DEFAULT_CONFIG.command_wait_time)),
        sample_count_per_read=int(os.environ.get("AQIMON_SAMPLE_COUNT_PER_READ", DEFAULT_CONFIG.sample_count_per_read)),
        server_port=int(os.environ.get("AQIMON_SERVER_PORT", DEFAULT_CONFIG.server_port)),
        server_host=os.environ.get("AQIMON_SERVER_HOST", DEFAULT_CONFIG.server_host),
        epa_lookback_minutes=int(os.environ.get("AQIMON_EPA_LOOKBACK_MIN", DEFAULT_CONFIG.epa_lookback_minutes)),
    )



"""Byte constants for the SDS011 device."""
from enum import Enum


# Message head constant
HEAD = b"\xaa"
# Message tail constant
TAIL = b"\xab"

# ID to send if the command should be issued to all sensor IDs.
ALL_SENSOR_ID = b"\xff\xff"

# The submit type
SUBMIT_TYPE = b"\xb4"


class ResponseType(Enum):
    """Response types for commands.

    GENERAL_RESPONSE is for all commands except query.
    QUERY_RESPONSE only applies to the query command.
    """

    GENERAL_RESPONSE = b"\xc5"
    # Query command has its own response type.
    QUERY_RESPONSE = b"\xc0"


class Command(Enum):
    """Possible commands for the device."""

    SET_REPORTING_MODE = b"\x02"
    QUERY = b"\x04"
    SET_DEVICE_ID = b"\x05"
    SET_SLEEP = b"\x06"
    SET_WORKING_PERIOD = b"\x08"
    CHECK_FIRMWARE_VERSION = b"\x07"


class OperationType(Enum):
    """Operation type for many commands.

    Many commands have two modes, one for setting a value, and another for retrieving.
    """

    QUERY = b"\x00"
    SET_MODE = b"\x01"


class ReportingMode(Enum):
    """Reporting mode for the device.

    ACTIVE mode means that the device is constantly returning read data from the device, and won't respond correctly
    to other query requests.

    QUERYING mode means that the device won't return read data unless explicitly asked for it.
    """

    ACTIVE = b"\x00"
    QUERYING = b"\x01"


class SleepState(Enum):
    """State of the device, either working or sleeping."""

    SLEEP = b"\x00"
    WORK = b"\x01"



"""Core server module."""
from fastapi import FastAPI, Depends
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi_utils.tasks import repeat_every
from starlette.responses import FileResponse
import uvicorn
import databases
from pathlib import Path
from datetime import datetime, timedelta
from .database import (
    get_all_reads,
    get_all_epa_aqis,
    add_read,
    add_epa_read,
    get_averaged_reads,
    create_tables,
    clean_old,
    ReadLogEntry,
    EpaAqiLogEntry,
)
from .read import AqiRead, Reader
from .read.mock import MockReader
from .read.novapm import OpinionatedReader
from . import aqi_common
from .config import Config, get_config_from_env
import logging
from functools import lru_cache
from dataclasses import dataclass
from typing import Optional, List

log = logging.getLogger(__name__)

app = FastAPI()

project_root = Path(__file__).parent.resolve()
static_dir = project_root / "static"
app.mount("/static", StaticFiles(directory=static_dir), name="static")

POLLUTANT_MAP = {aqi_common.Pollutant.PM_10: "PM10", aqi_common.Pollutant.PM_25: "PM25"}


@dataclass
class ScheduledReader:
    """Simple state wrapper to track reader, and next schedule time."""

    next_schedule: Optional[datetime]
    reader: Reader


@lru_cache(maxsize=1)
def get_config():
    """Retrieve config from environment.

    Uses LRU cache to simulate a singleton.
    """
    return get_config_from_env()


@lru_cache(maxsize=1)
def get_database(config: Config = Depends(get_config)) -> databases.Database:
    """Retrieve database instance.

    Creates a path to the database if it doesn't exist.

    Uses LRU cache to simulate a singleton.
    """
    Path(config.database_path).parent.mkdir(parents=True, exist_ok=True)
    return databases.Database(f"sqlite+aiosqlite:///{config.database_path}")


def build_reader() -> ScheduledReader:
    """Retrieve the reader class.

    Uses LRU cache to simulate a singleton.
    """
    conf = get_config()
    if conf.reader_type == "MOCK":
        return ScheduledReader(None, MockReader())
    elif conf.reader_type == "NOVAPM":
        return ScheduledReader(
            None,
            OpinionatedReader(
                ser_dev=conf.usb_path,
                warm_up_secs=conf.warm_up_sec,
                iterations=conf.sample_count_per_read,
                sleep_time=conf.sleep_sec_between_reads,
                command_wait_time=conf.command_wait_time,
            ),
        )
    else:
        raise Exception("Invalid reader type specified")


reader = build_reader()


def get_reader() -> ScheduledReader:
    """Retrieve the global reader to find state from.

    TODO: Change this to not rely on global module-level state.
    """
    global reader
    return reader


@app.on_event("startup")
async def database_connect():
    """Connect to the database, and create tables on startup."""
    database = get_database(get_config_from_env())
    await database.connect()
    await create_tables(database)


@app.on_event("shutdown")
async def database_disconnect():
    """Disconnect from the database on shutdown."""
    database = get_database(get_config_from_env())
    await database.disconnect()


@app.on_event("startup")
async def read_from_device() -> None:
    """Background cron task to read from the device."""
    config = get_config_from_env()
    database = get_database(config)
    scheduled_reader = get_reader()

    async def read_function() -> None:
        try:
            # Set the approximate time of the next read
            result: AqiRead = await scheduled_reader.reader.read()
            event_time = datetime.now()
            await add_read(
                dbconn=database,
                event_time=event_time,
                pm25=result.pmtwofive,
                pm10=result.pmten,
            )

            averaged_reads = await get_averaged_reads(
                dbconn=database, lookback_to=event_time - timedelta(minutes=config.epa_lookback_minutes)
            )
            if averaged_reads:
                read_list = [
                    aqi_common.PollutantReading(averaged_reads.avg_pm25, aqi_common.Pollutant.PM_25),
                    aqi_common.PollutantReading(averaged_reads.avg_pm10, aqi_common.Pollutant.PM_10),
                ]
                epa_aqi = aqi_common.calculate_epa_aqi(read_list)

                if epa_aqi:
                    pollutant = POLLUTANT_MAP.get(epa_aqi.responsible_pollutant)

                    if pollutant is None:
                        raise Exception(f"Invalid Pollutant! {epa_aqi.responsible_pollutant}")

                    await add_epa_read(
                        dbconn=database,
                        event_time=event_time,
                        epa_aqi=epa_aqi.reading,
                        pollutant=pollutant,
                        read_count=averaged_reads.count,
                        oldest_read_time=averaged_reads.oldest_read_time,
                    )
                else:
                    log.warning("No EPA Value was calculated.")

            await clean_old(dbconn=database, retention_minutes=config.retention_minutes)
            scheduled_reader.next_schedule = datetime.now() + timedelta(seconds=config.poll_frequency_sec)
        except Exception as e:
            log.exception("Failed to retrieve data from reader", e)
            scheduled_reader.next_schedule = datetime.now() + timedelta(seconds=config.poll_frequency_sec)

    # Note that we leverage the @repeat_every decorator here, but as a regular function call.  This allows us to
    # use a non-global config object to specify the poll frequency
    repeater = repeat_every(seconds=config.poll_frequency_sec)
    await repeater(read_function)()


def convert_all_to_view_dict(reads: List[ReadLogEntry], epas: List[EpaAqiLogEntry]):
    """Convert data result to dictionary for view."""
    view = {
        "reads": [{"t": int(x.event_time.timestamp()), "pm25": x.pm25, "pm10": x.pm10} for x in reads],
        "epas": [{"t": int(x.event_time.timestamp()), "epa": x.epa_aqi} for x in epas],
    }
    return view


@app.get("/", response_class=HTMLResponse)
async def home():
    """Return the index page."""
    return FileResponse(static_dir / "index.html")


@app.get("/api/sensor_data")
async def all_data(
    database: databases.Database = Depends(get_database),
    window: str = "all",
):
    """Retrieve sensor data for the given window."""
    window_delta = None
    if window == "hour":
        window_delta = timedelta(hours=1)
    elif window == "day":
        window_delta = timedelta(days=1)
    elif window == "week":
        window_delta = timedelta(weeks=1)
    if window_delta:
        all_reads = await get_all_reads(database, datetime.now() - window_delta)
        all_epas = await get_all_epa_aqis(database, datetime.now() - window_delta)
    else:
        all_reads = await get_all_reads(database, None)
        all_epas = await get_all_epa_aqis(database, None)
    all_json = convert_all_to_view_dict(all_reads, all_epas)
    return all_json


@app.get("/api/status")
async def status(reader: ScheduledReader = Depends(get_reader)):
    """Get the system status."""
    last_exception = reader.reader.get_state().last_exception

    return {
        "reader_status": str(reader.reader.get_state().status.name),
        "reader_exception": str(last_exception) if last_exception else None,
        "next_schedule": int(reader.next_schedule.timestamp() * 1000) if reader.next_schedule else None,
    }


def start():
    """Start the server."""
    env_config = get_config_from_env()
    uvicorn.run(app, host=env_config.server_host, port=env_config.server_port)


def debug():
    """Start the server in debug mode, with hotswapping code."""
    env_config = get_config_from_env()
    uvicorn.run("aqimon.server:app", host=env_config.server_host, port=env_config.server_port, reload=True)



"""Mock reader class.

Useful for testing locally, when you don't have a real reader available.
"""
import random
import asyncio
import logging
from aqimon.read import AqiRead, ReaderState, ReaderStatus


class MockReader:
    """Mock reader class."""

    def __init__(self, fake_sleep_secs=5, raise_error_odds=50):
        """Create a mock reader."""
        self.state = ReaderState(ReaderStatus.IDLE, None)
        self.fake_sleep_secs = fake_sleep_secs
        self.raise_error_odds = raise_error_odds

    async def read(self) -> AqiRead:
        """Read from the 'device'.

        Returns randomized data for the mock class.

        Also, randomly fails some percentage of the time.
        """
        try:
            self.state = ReaderState(ReaderStatus.READING, None)
            raise_error_roll = random.randint(0, 100)
            if raise_error_roll < self.raise_error_odds:
                raise Exception("Fake error from the reader.")
            pm25: float = round(random.uniform(0.0, 500.4), 2)
            pm10: float = round(random.uniform(0.0, 300.0), 2)
            result = AqiRead(pmtwofive=pm25, pmten=pm10)

            await asyncio.sleep(self.fake_sleep_secs)
            self.state = ReaderState(ReaderStatus.IDLE, None)
            return result
        except Exception as e:
            self.state = ReaderState(ReaderStatus.ERRORING, e)
            raise e

    def get_state(self) -> ReaderState:
        """Get the state of the reader."""
        logging.info(self.state)
        return self.state



"""Reader module."""
from typing import Protocol, Optional
from dataclasses import dataclass
from enum import Enum


class ReaderStatus(Enum):
    """Enum of possible reader states."""

    IDLE = 1
    WARM_UP = 2
    READING = 3
    ERRORING = 4


@dataclass(frozen=True)
class ReaderState:
    """State of a reader.

    Used to report status and errors to the user.
    """

    status: ReaderStatus
    last_exception: Optional[Exception]


@dataclass(frozen=True)
class AqiRead:
    """A raw read from an AQI device."""

    pmtwofive: float
    pmten: float


class Reader(Protocol):
    """Protocol for a reader.

    Hypothetically could be implemented for any number of devices.
    """

    async def read(self) -> AqiRead:
        """Read from the device."""
        pass

    def get_state(self) -> ReaderState:
        """Get the state of the reader."""
        pass



"""Nova PM SDS011 Reader module.

Device: https://www.amazon.com/SDS011-Quality-Detection-Conditioning-Monitor/dp/B07FSDMRR5

Spec: https://cdn.sparkfun.com/assets/parts/1/2/2/7/5/Laser_Dust_Sensor_Control_Protocol_V1.3.pdf
Spec: https://cdn-reichelt.de/documents/datenblatt/X200/SDS011-DATASHEET.pdf
"""

import serial
import time
from .responses import (
    QueryReadResponse,
    ReportingModeReadResponse,
    SleepWakeReadResponse,
    DeviceIdResponse,
    CheckFirmwareReadResponse,
    WorkingPeriodReadResponse,
)
from . import constants as con
from .exceptions import IncompleteReadException, IncorrectCommandException, IncorrectCommandCodeException


class NovaPmReader:
    """NOVA PM SDS011 Reader."""

    def __init__(self, ser_dev: serial.Serial, send_command_sleep: int = 1):
        """Create the device."""
        self.ser = ser_dev
        self.send_command_sleep = send_command_sleep

    def request_data(self) -> None:
        """Request device to return pollutant data."""
        cmd = con.Command.QUERY.value + (b"\x00" * 12) + con.ALL_SENSOR_ID
        self._send_command(cmd)

    def query_data(self) -> QueryReadResponse:
        """Query the device for pollutant data."""
        return QueryReadResponse(self._read_response())

    def request_reporting_mode(self) -> None:
        """Request device to return the current reporting mode."""
        cmd = (
            con.Command.SET_REPORTING_MODE.value
            + con.OperationType.QUERY.value
            + con.ReportingMode.ACTIVE.value
            + (b"\x00" * 10)
            + con.ALL_SENSOR_ID
        )
        self._send_command(cmd)

    def query_reporting_mode(self) -> ReportingModeReadResponse:
        """Get the current reporting mode of the device."""
        return ReportingModeReadResponse(self._read_response())

    def set_active_mode(self) -> None:
        """Set the reporting mode to active."""
        self._set_reporting_mode(con.ReportingMode.ACTIVE)
        try:
            self.query_reporting_mode()
        except IncorrectCommandException:
            pass
        except IncompleteReadException:
            pass

    def set_query_mode(self) -> None:
        """Set the reporting mode to querying."""
        self._set_reporting_mode(con.ReportingMode.QUERYING)
        try:
            self.query_reporting_mode()
        except IncorrectCommandException:
            pass
        except IncompleteReadException:
            pass
        except IncorrectCommandCodeException:
            pass

    def _set_reporting_mode(self, reporting_mode: con.ReportingMode) -> None:
        """Set the reporting mode, either ACTIVE or QUERYING.

        ACTIVE mode means the device will always return a Query command response when data is asked for, regardless of
        what command was sent.

        QUERYING mode means the device will only return responses to submitted commands, even for Query commands.

        ACTIVE mode is the factory default, but generally, QUERYING mode is preferrable for the longevity of the device.
        """
        cmd = (
            con.Command.SET_REPORTING_MODE.value
            + con.OperationType.SET_MODE.value
            + reporting_mode.value
            + (b"\x00" * 10)
            + con.ALL_SENSOR_ID
        )
        self._send_command(cmd)
        # Switching between reporting modes is finicky; resetting the serial connection seems to address issues.
        self.ser.close()
        self.ser.open()

    def request_sleep_state(self) -> None:
        """Get the current sleep state."""
        cmd = con.Command.SET_SLEEP.value + con.OperationType.QUERY.value + b"\x00" + (b"\x00" * 10) + con.ALL_SENSOR_ID
        self._send_command(cmd)

    def query_sleep_state(self) -> SleepWakeReadResponse:
        """Get the current sleep state."""
        return SleepWakeReadResponse(self._read_response())

    def set_sleep_state(self, sleep_state: con.SleepState) -> None:
        """Set the sleep state, either wake or sleep."""
        cmd = (
            con.Command.SET_SLEEP.value
            + con.OperationType.SET_MODE.value
            + sleep_state.value
            + (b"\x00" * 10)
            + con.ALL_SENSOR_ID
        )
        self._send_command(cmd)

    def sleep(self) -> None:
        """Put the device to sleep, turning off fan and diode."""
        self.set_sleep_state(con.SleepState.SLEEP)

    def wake(self) -> None:
        """Wake the device up to start reading."""
        self.set_sleep_state(con.SleepState.WORK)

    def safe_wake(self) -> None:
        """Wake the device up, if you don't know what mode its in.

        This operates as a fire-and-forget, even in query mode.  You shouldn't have to (and can't) query for a response
        after this command.
        """
        self.wake()
        # If we were in query mode, this would flush out the response.  If in active mode, this would be return read
        # data, but we don't care.
        self.ser.read(10)

    def set_device_id(self, device_id: bytes, target_device_id: bytes = con.ALL_SENSOR_ID) -> None:
        """Set the device ID."""
        if len(device_id) != 2 or len(target_device_id) != 2:
            raise AttributeError(f"Device ID must be 4 bytes, found {len(device_id)}, and {len(target_device_id)}")
        cmd = con.Command.SET_DEVICE_ID.value + (b"\x00" * 10) + device_id + target_device_id
        self._send_command(cmd)

    def query_device_id(self) -> DeviceIdResponse:
        """Set the device ID."""
        return DeviceIdResponse(self._read_response())

    def request_working_period(self) -> None:
        """Retrieve the current working period for the device."""
        cmd = con.Command.SET_WORKING_PERIOD.value + con.OperationType.QUERY.value + (b"\x00" * 11) + con.ALL_SENSOR_ID
        self._send_command(cmd)

    def query_working_period(self) -> WorkingPeriodReadResponse:
        """Retrieve the current working period for the device."""
        return WorkingPeriodReadResponse(self._read_response())

    def set_working_period(self, working_period: int) -> None:
        """Set the working period for the device.

        Working period must be between 0 and 30.

        0 means the device will read continuously.
        Any value 1-30 means the device will wake and read for 30 seconds every n*60-30 seconds.
        """
        if 0 >= working_period >= 30:
            raise AttributeError("Working period must be between 0 and 30")
        cmd = (
            con.Command.SET_WORKING_PERIOD.value
            + con.OperationType.SET_MODE.value
            + bytes([working_period])
            + (b"\x00" * 10)
            + con.ALL_SENSOR_ID
        )
        self._send_command(cmd)

    def request_firmware_version(self) -> None:
        """Retrieve the firmware version from the device."""
        cmd = con.Command.CHECK_FIRMWARE_VERSION.value + (b"\x00" * 12) + con.ALL_SENSOR_ID
        self._send_command(cmd)

    def query_firmware_version(self) -> CheckFirmwareReadResponse:
        """Retrieve the firmware version from the device."""
        return CheckFirmwareReadResponse(self._read_response())

    def _send_command(self, cmd: bytes):
        """Send a command to the device as bytes."""
        head = con.HEAD + con.SUBMIT_TYPE
        full_command = head + cmd + bytes([self._cmd_checksum(cmd)]) + con.TAIL
        if len(full_command) != 19:
            raise Exception(f"Command length must be 19, but was {len(full_command)}")
        self.ser.write(full_command)
        time.sleep(self.send_command_sleep)

    def _read_response(self) -> bytes:
        """Read a response from the device."""
        result = self.ser.read(10)
        if len(result) != 10:
            raise IncompleteReadException(len(result))
        return result

    def _cmd_checksum(self, data: bytes) -> int:
        """Generate a checksum for the data bytes of a command."""
        if len(data) != 15:
            raise AttributeError("Invalid checksum length.")
        return sum(d for d in data) % 256


class QueryModeReader:
    """Reader working in query mode."""

    def __init__(self, ser_dev: serial.Serial, send_command_sleep: int = 1):
        """Create the device."""
        self.base_reader = NovaPmReader(ser_dev=ser_dev, send_command_sleep=send_command_sleep)
        self.base_reader.safe_wake()
        self.base_reader.set_query_mode()

    def query(self) -> QueryReadResponse:
        """Query the device for pollutant data."""
        self.base_reader.request_data()
        return self.base_reader.query_data()

    def get_reporting_mode(self) -> ReportingModeReadResponse:
        """Get the current reporting mode of the device."""
        self.base_reader.request_reporting_mode()
        return self.base_reader.query_reporting_mode()

    def get_sleep_state(self) -> SleepWakeReadResponse:
        """Get the current sleep state."""
        self.base_reader.request_sleep_state()
        return self.base_reader.query_sleep_state()

    def sleep(self) -> SleepWakeReadResponse:
        """Put the device to sleep, turning off fan and diode."""
        self.base_reader.sleep()
        return self.base_reader.query_sleep_state()

    def wake(self) -> SleepWakeReadResponse:
        """Wake the device up to start reading."""
        self.base_reader.wake()
        return self.base_reader.query_sleep_state()

    def set_device_id(self, device_id: bytes, target_device_id: bytes = con.ALL_SENSOR_ID) -> DeviceIdResponse:
        """Set the device ID."""
        self.base_reader.set_device_id(device_id, target_device_id)
        return self.base_reader.query_device_id()

    def get_working_period(self) -> WorkingPeriodReadResponse:
        """Retrieve the current working period for the device."""
        self.base_reader.request_working_period()
        return self.base_reader.query_working_period()

    def set_working_period(self, working_period: int) -> WorkingPeriodReadResponse:
        """Set the working period for the device.

        Working period must be between 0 and 30.

        0 means the device will read continuously.
        Any value 1-30 means the device will wake and read for 30 seconds every n*60-30 seconds.
        """
        self.base_reader.set_working_period(working_period)
        return self.base_reader.query_working_period()

    def get_firmware_version(self) -> CheckFirmwareReadResponse:
        """Retrieve the firmware version from the device."""
        self.base_reader.request_firmware_version()
        return self.base_reader.query_firmware_version()


class ActiveModeReader:
    """Active Mode Reader.

    Use with caution! Active mode is unpredictable.  Query mode is much preferred.
    """

    def __init__(self, ser_dev: serial.Serial, send_command_sleep: int = 2):
        """Create the device."""
        self.base_reader = NovaPmReader(ser_dev=ser_dev, send_command_sleep=send_command_sleep)
        self.ser_dev = ser_dev
        self.base_reader.safe_wake()
        self.base_reader.set_active_mode()

    def query(self) -> QueryReadResponse:
        """Query the device for pollutant data."""
        return self.base_reader.query_data()

    def sleep(self) -> None:
        """Put the device to sleep, turning off fan and diode."""
        self.base_reader.sleep()

        # Sleep seems to behave very strangely in active mode.  It continually outputs data for old commands for quite
        # a while before eventually having nothing to report.  This forces it to "drain" whatever it was doing before
        # returning, but also feels quite dangerous.
        while len(self.ser_dev.read(10)) == 10:
            pass

    def wake(self) -> None:
        """Wake the device up to start reading."""
        self.base_reader.wake()
        self.ser_dev.read(10)

    def set_device_id(self, device_id: bytes, target_device_id: bytes = con.ALL_SENSOR_ID) -> None:
        """Set the device ID."""
        self.base_reader.set_device_id(device_id, target_device_id)
        self.ser_dev.read(10)

    def set_working_period(self, working_period: int) -> None:
        """Set the working period for the device.

        Working period must be between 0 and 30.

        0 means the device will read continuously.
        Any value 1-30 means the device will wake and read for 30 seconds every n*60-30 seconds.
        """
        self.base_reader.set_working_period(working_period)
        self.ser_dev.read(10)



"""Common AQI functionality.

Ranges here are ripped from public sites.
"""

from typing import List, Tuple, Optional, Dict
from enum import Enum
import dataclasses


class Pollutant(Enum):
    """Enum of possible pollutans."""

    PM_25 = 0
    PM_10 = 1


AQI: List[Tuple[int, int]] = [
    (0, 50),
    (51, 100),
    (101, 150),
    (151, 200),
    (201, 300),
    (301, 400),
    (401, 500),
]
PM_25: List[Tuple[float, float]] = [
    (0.0, 12.0),
    (12.1, 35.4),
    (35.5, 55.4),
    (55.5, 150.4),
    (150.5, 250.4),
    (250.5, 350.4),
    (350.5, 500.4),
]

PM_10: List[Tuple[float, float]] = [
    (0, 54),
    (55, 154),
    (155, 254),
    (255, 354),
    (355, 424),
    (425, 504),
    (505, 604),
]

AQI_LOOKUP_MAP: Dict[Pollutant, List[Tuple[float, float]]] = {Pollutant.PM_25: PM_25, Pollutant.PM_10: PM_10}


@dataclasses.dataclass(frozen=True)
class PollutantReading:
    """A reading for a given pollutant."""

    reading: float
    pollutant: Pollutant


@dataclasses.dataclass(frozen=True)
class EpaAqi:
    """An EPA AQI value, with the pollutant responsible for the value."""

    reading: float
    responsible_pollutant: Pollutant


def get_level_from_pm25(pm25: float) -> int:
    """Get the EPA level from a PM25 reading."""
    for i, pair in enumerate(PM_25):
        if pair[0] <= pm25 <= pair[1]:
            return i
    raise ValueError("Invalid PM value")


def calculate_epa_aqi(readings: List[PollutantReading]) -> Optional[EpaAqi]:
    """Calculate the EPA AQI from a list of pollutant readings.

    The worst possible value will be reported.
    """
    max_value: Optional[EpaAqi] = None
    for reading in readings:
        epa_value = calculate_epa_aqi_raw(reading)
        if max_value is None or max_value.reading < epa_value:
            max_value = EpaAqi(epa_value, reading.pollutant)
    return max_value


def calculate_epa_aqi_raw(pollutant_reading: PollutantReading) -> int:
    """Calculate the EPA AQI based on a pollutant reading."""
    ranges = AQI_LOOKUP_MAP[pollutant_reading.pollutant]
    for i, pm_range in enumerate(ranges):
        if pm_range[0] <= pollutant_reading.reading <= pm_range[1]:
            aqi_low = AQI[i][0]
            aqi_high = AQI[i][1]
            pm_low = pm_range[0]
            pm_high = pm_range[1]
            epa = ((aqi_high - aqi_low) / (pm_high - pm_low)) * (pollutant_reading.reading - pm_low) + aqi_low
            return round(epa)
    return -1



"""Nova PM SDS011 Reader module.

https://www.amazon.com/SDS011-Quality-Detection-Conditioning-Monitor/dp/B07FSDMRR5
"""
import asyncio

from . import AqiRead, ReaderState, ReaderStatus
import serial
from typing import Union
from .sds011 import QueryModeReader
from statistics import mean


class OpinionatedReader:
    """NOVA PM SDS011 Reader."""

    def __init__(
        self,
        ser_dev: Union[str, serial.Serial],
        warm_up_secs: int = 15,
        iterations: int = 5,
        sleep_time: int = 3,
        command_wait_time: int = 15,
    ):
        """Create the device."""
        if isinstance(ser_dev, str):
            ser_dev = serial.Serial(ser_dev, timeout=2)

        self.reader = QueryModeReader(ser_dev=ser_dev, send_command_sleep=command_wait_time)

        # Initial the reader to be in the mode we want.
        self.reader.wake()
        self.reader.set_working_period(0)

        self.warm_up_secs = warm_up_secs
        self.iterations = iterations
        self.sleep_time = sleep_time

        self.state = ReaderState(ReaderStatus.IDLE, None)

    async def read(self) -> AqiRead:
        """Read from the device."""
        try:
            self.reader.wake()
            self.state = ReaderState(ReaderStatus.WARM_UP, None)
            await asyncio.sleep(self.warm_up_secs)
            self.state = ReaderState(ReaderStatus.READING, None)
            pm25_reads = []
            pm10_reads = []
            for x in range(0, self.iterations):
                await asyncio.sleep(self.sleep_time)
                result = self.reader.query()
                pm25_reads.append(result.pm25)
                pm10_reads.append(result.pm10)
            self.reader.sleep()
            self.state = ReaderState(ReaderStatus.IDLE, None)
            return AqiRead(pmtwofive=mean(pm25_reads), pmten=mean(pm10_reads))
        except Exception as e:
            self.state = ReaderState(ReaderStatus.ERRORING, e)
            self.reader.sleep()
            raise e

    def get_state(self) -> ReaderState:
        """Get the current state of the reader."""
        return self.state



"""Response objects for SDS011.

Creates and validates typed classes from binary responses from the device.
"""
from .constants import (
    HEAD,
    TAIL,
    Command,
    ResponseType,
    SleepState,
    OperationType,
    ReportingMode,
)
from .exceptions import (
    ChecksumFailedException,
    IncorrectCommandException,
    IncorrectCommandCodeException,
    IncorrectWrapperException,
    IncompleteReadException,
)


class ReadResponse:
    """Generic read response object for responses from SDS011."""

    def __init__(self, data: bytes, command_code: Command, response_type: ResponseType = ResponseType.GENERAL_RESPONSE):
        """Create a read response."""
        if len(data) != 10:
            raise IncompleteReadException()

        self.head = data[0:1]
        self.cmd_id = data[1:2]
        self.data = data[2:8]
        self.device_id = data[6:8]
        self.checksum: int = data[8]
        self.tail = data[9:10]
        self.expected_command_code = command_code
        self.expected_response_type = response_type
        # Check it!
        self.verify()

    def verify(self):
        """Verify the read data."""
        if self.head != HEAD:
            raise IncorrectWrapperException()
        if self.tail != TAIL:
            raise IncorrectWrapperException()
        if self.checksum != self.calc_checksum():
            raise ChecksumFailedException(expected=self.checksum, actual=self.calc_checksum())
        if self.cmd_id != self.expected_response_type.value:
            raise IncorrectCommandException(expected=self.expected_response_type.value, actual=self.cmd_id)

        # Query responses don't validate the command code
        if (
            self.expected_response_type != ResponseType.QUERY_RESPONSE
            and bytes([self.data[0]]) != self.expected_command_code.value
        ):
            raise IncorrectCommandCodeException(expected=self.expected_command_code.value, actual=self.data[0])

    def calc_checksum(self) -> int:
        """Calculate the checksum for the read data."""
        return sum(d for d in self.data) % 256


class QueryReadResponse(ReadResponse):
    """Query read response."""

    def __init__(self, data: bytes):
        """Create a query read response."""
        super().__init__(data, command_code=Command.QUERY, response_type=ResponseType.QUERY_RESPONSE)

        self.pm25: float = int.from_bytes(data[2:4], byteorder="little") / 10
        self.pm10: float = int.from_bytes(data[4:6], byteorder="little") / 10


class ReportingModeReadResponse(ReadResponse):
    """Reporting mode response."""

    def __init__(self, data: bytes):
        """Create a reporting mode response."""
        super().__init__(data, command_code=Command.SET_REPORTING_MODE)
        self.operation_type = OperationType(self.data[1:2])
        self.state = ReportingMode(self.data[2:3])


class DeviceIdResponse(ReadResponse):
    """Device ID response."""

    def __init__(self, data: bytes):
        """Create a device ID response."""
        super().__init__(data, command_code=Command.SET_DEVICE_ID)


class SleepWakeReadResponse(ReadResponse):
    """Sleep/Wake Response."""

    def __init__(self, data: bytes):
        """Create a sleep/wake response."""
        super().__init__(data, command_code=Command.SET_SLEEP)
        self.operation_type = OperationType(self.data[1:2])
        self.state = SleepState(self.data[2:3])


class WorkingPeriodReadResponse(ReadResponse):
    """Working period response."""

    def __init__(self, data: bytes):
        """Create a working period response."""
        super().__init__(data, command_code=Command.SET_WORKING_PERIOD)
        self.operation_type = OperationType(self.data[1:2])
        self.interval: int = self.data[2]


class CheckFirmwareReadResponse(ReadResponse):
    """Firmware response."""

    def __init__(self, data: bytes):
        """Create a firmware response."""
        super().__init__(data, command_code=Command.CHECK_FIRMWARE_VERSION)
        self.year = self.data[1]
        self.month = self.data[2]
        self.day = self.data[3]
